/*******************************************************************************
 * Copyright (c) 2004, 2005 Actuate Corporation.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *  Actuate Corporation  - initial API and implementation
 *******************************************************************************/

package org.eclipse.birt.data.engine.aggregation;

import java.util.HashMap;

import org.eclipse.birt.core.framework.FrameworkException;
import org.eclipse.birt.core.framework.IConfigurationElement;
import org.eclipse.birt.core.framework.IExtension;
import org.eclipse.birt.core.framework.IExtensionPoint;
import org.eclipse.birt.core.framework.IExtensionRegistry;
import org.eclipse.birt.core.framework.Platform;
import org.eclipse.birt.data.engine.api.aggregation.IAggregation;
import org.eclipse.birt.data.engine.api.aggregation.IBuildInAggregation;
import org.eclipse.birt.data.engine.core.DataException;
import org.eclipse.birt.data.engine.i18n.ResourceConstants;


/**
 * The instances of this class manage all aggregations, including BuiltInAggregation (as defined
 * in BuiltInAggregationFactory) and customized aggregation (as defined in plugins).
 */

public class AggregationFactory 
{
	private static final String ATTRIBUTE_AGGREGATION_CLASS = "aggregationClass";
	private static final String ATTRIBUTE_AGGREGATION_NAME = "name";
	private static final String EXTENSION_POINT = "org.eclipse.birt.data.aggregation";
	private static final String ELEMENT_AGGREGATIONS = "Aggregations";
	private static final String ELEMENT_AGGREGATION = "Aggregation";
	
	private static AggregationFactory instance = null;
	private static HashMap aggregations = null;
	private static HashMap aggrAdapterMap = null;


	/**
	 * Return a shared instance of AggregationFactory.
	 * @return
	 * @throws DataException
	 */
	public static AggregationFactory getInstance( ) throws DataException
	{
		if ( instance == null )
		{
			synchronized ( AggregationFactory.class )
			{
				if ( instance == null )
					instance = new AggregationFactory( );
			}
		}
		
		return instance;
	}
	
	/**
	 * Destroy shared instance of AggregationFactory.
	 * 
	 */
	public static void destroyInstance( )
	{
		instance = null;
		aggregations = null;
		aggrAdapterMap = null;
	}

	
	private AggregationFactory( ) throws DataException
	{
		populateAggregations( );
	}

	/**
	 * Populate the aggregations. Put all aggregations to the HashMap "aggregations" 
	 * @throws DataException
	 */
	private void populateAggregations( ) throws DataException
	{
		aggregations = new HashMap();
		aggrAdapterMap = new HashMap();
		populateAggrAdapterMap( );
		populateBuiltInAggregations( );
		populateExtendedAggregations( );
	}

	/**
	 * Populate model aggregation names to build in aggregation names.
	 * Note: This adapter should be synchronized with DataAdapterUtil in Model.
	 */
	private static void populateAggrAdapterMap( )
	{
		aggrAdapterMap.put( "sum", IBuildInAggregation.TOTAL_SUM_FUNC );
		aggrAdapterMap.put( "count", IBuildInAggregation.TOTAL_COUNT_FUNC );
		aggrAdapterMap.put( "min", IBuildInAggregation.TOTAL_MIN_FUNC );
		aggrAdapterMap.put( "max", IBuildInAggregation.TOTAL_MAX_FUNC );
		aggrAdapterMap.put( "average", IBuildInAggregation.TOTAL_AVE_FUNC );
		aggrAdapterMap.put( "weighted-avg", IBuildInAggregation.TOTAL_WEIGHTEDAVE_FUNC );
		aggrAdapterMap.put( "stddev", IBuildInAggregation.TOTAL_STDDEV_FUNC );
		aggrAdapterMap.put( "first", IBuildInAggregation.TOTAL_FIRST_FUNC );
		aggrAdapterMap.put( "last", IBuildInAggregation.TOTAL_LAST_FUNC );
		aggrAdapterMap.put( "mode", IBuildInAggregation.TOTAL_MODE_FUNC );
		aggrAdapterMap.put( "moving-ave", IBuildInAggregation.TOTAL_MOVINGAVE_FUNC );
		aggrAdapterMap.put( "median", IBuildInAggregation.TOTAL_MEDIAN_FUNC );
		aggrAdapterMap.put( "variance", IBuildInAggregation.TOTAL_VARIANCE_FUNC );
		aggrAdapterMap.put( "running-sum", IBuildInAggregation.TOTAL_RUNNINGSUM_FUNC );
		aggrAdapterMap.put( "irr", IBuildInAggregation.TOTAL_IRR_FUNC );
		aggrAdapterMap.put( "mirr", IBuildInAggregation.TOTAL_MIRR_FUNC );
		aggrAdapterMap.put( "npv", IBuildInAggregation.TOTAL_NPV_FUNC );
		aggrAdapterMap.put( "running-npv", IBuildInAggregation.TOTAL_RUNNINGNPV_FUNC );
		aggrAdapterMap.put( "count-distinct", IBuildInAggregation.TOTAL_COUNTDISTINCT_FUNC );
		aggrAdapterMap.put( "running-count", IBuildInAggregation.TOTAL_RUNNINGCOUNT_FUNC );
		aggrAdapterMap.put( "is-top-n", IBuildInAggregation.TOTAL_TOP_N_FUNC );
		aggrAdapterMap.put( "is-bottom-n", IBuildInAggregation.TOTAL_BOTTOM_N_FUNC );
		aggrAdapterMap.put( "is-top-n-percent", IBuildInAggregation.TOTAL_TOP_PERCENT_FUNC );
		aggrAdapterMap.put( "is-bottom-n-percent", IBuildInAggregation.TOTAL_BOTTOM_PERCENT_FUNC );
		aggrAdapterMap.put( "percent-rank", IBuildInAggregation.TOTAL_PERCENT_RANK_FUNC );
		aggrAdapterMap.put( "percentile", IBuildInAggregation.TOTAL_PERCENTILE_FUNC );
		aggrAdapterMap.put( "quartile", IBuildInAggregation.TOTAL_QUARTILE_FUNC );
		aggrAdapterMap.put( "percent-sum", IBuildInAggregation.TOTAL_PERCENTSUM_FUNC );
		aggrAdapterMap.put( "rank", IBuildInAggregation.TOTAL_RANK_FUNC );
	}
	
	/**
	 * populate the built-in Aggregations
	 */
	private void populateBuiltInAggregations( )
	{
		String[] builtInAggrNames = BuiltInAggregationFactory.getInstance( )
				.getAggregationNames( );
		for ( int i = 0; i < builtInAggrNames.length; i++ )
		{
			aggregations.put( builtInAggrNames[i],
					BuiltInAggregationFactory.getInstance( )
							.getAggregation( builtInAggrNames[i] ) );
		}
	}

	/**
	 * Load aggregation extensions from runtime environment.
	 * @throws DataException 
	 * @throws FrameworkException 
	 */
	private void populateExtendedAggregations( ) throws DataException 
	{
		IExtensionRegistry extReg = Platform.getExtensionRegistry( );

		IExtensionPoint extPoint = extReg.getExtensionPoint( EXTENSION_POINT );

		if ( extPoint == null )
			return;

		IExtension[] exts = extPoint.getExtensions( );
		if ( exts == null )
			return;

		for ( int e = 0; e < exts.length; e++ )
		{
			IConfigurationElement[] configElems = exts[e].getConfigurationElements( );
			if ( configElems == null )
				continue;

			for ( int i = 0; i < configElems.length; i++ )
			{
				if ( configElems[i].getName( ).equals( ELEMENT_AGGREGATIONS ) )
				{
					IConfigurationElement[] subElems = configElems[i].getChildren( ELEMENT_AGGREGATION );
					for ( int j = 0; j < subElems.length; j++ )
					{

						String name = subElems[j].getAttribute( ATTRIBUTE_AGGREGATION_NAME );

						try
						{
							if ( aggregations.put( name.toUpperCase( ).trim( ),
									subElems[j].createExecutableExtension( ATTRIBUTE_AGGREGATION_CLASS ) ) != null )
								throw new DataException( ResourceConstants.DUPLICATE_AGGREGATION_NAME,
										name );
						}
						catch ( FrameworkException exception )
						{
							DataException ep = new DataException( ResourceConstants.CANNOT_INSTANTIATE_AGGREGATION_FACTORY,
									name );
							throw ep;
						}
					}
				}
			}
		}
	}
	
	/**
	 * Return an IAggregation instance according to the given aggregation name
	 * @param name the name of the aggregation
	 * @return an IAggregation instance specified by the give named
	 */
	public IAggregation getAggregation( String name )
	{
		assert name != null;
		// try to fetch the aggregation as build-in aggregation name
		IAggregation aggr = (IAggregation) aggregations.get( name.toUpperCase( ) );
		// if fails, it may be an aggregation name used in model and will be adapted to
		// build-in aggregation name
		if ( aggr == null )
		{
			String buildInAggrName = (String) aggrAdapterMap.get( name );
			aggr = (IAggregation) aggregations.get( buildInAggrName );
		}
		return aggr;
	}
}
