/*******************************************************************************
 * Copyright (c) 2004, 2005 Actuate Corporation.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *  Actuate Corporation  - initial API and implementation
 *******************************************************************************/
package org.eclipse.birt.data.engine.olap.cursor;

import java.io.IOException;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.Vector;

import javax.olap.OLAPException;

import org.eclipse.birt.data.engine.i18n.ResourceConstants;
import org.eclipse.birt.data.engine.olap.data.api.IAggregationResultSet;
import org.eclipse.birt.data.engine.olap.data.api.IDimensionSortDefn;
import org.eclipse.birt.data.engine.olap.driver.DimensionAxis;

/**
 * edgeInfoArray[axisIndex] is an Vector of EdgeInfo objects at the specified axis index.
 * axisIndex is a 0-based index.
 * Example: 
 * Dim  Country 	City 	    Product 	
 * 0: 	CHINA 		BEIJING 	P1 
 * 1: 	CHINA		BEIJING		P2 
 * 2: 	CHINA		BEIJING		P3
 * 3: 	CHINA		SHANGHAI	P1 
 * 4:	CHINA		SHANGHAI	P2
 * 5: 	CHINA 		SHANGHAI	P3
 * 6: 	USA 		CHICAGO		P1
 * 7: 	USA			NEW YORK	P1
 * 8: 	USA 		NEW YORK	P2
 * 
 * edgeInfo: (start, end) 
 * 		Country		    City 		    Product
 * ============================================ 
 * 0: 	-1,0 			0,0 			0,0
 * 1: 	-1,2 			0,3 			0,1 
 * 2:	 				1,6 			0,2 
 * 3: 					1,7 			1,3 
 * 4: 									1,4 
 * 5: 									1,5  
 * 6:                                   2,6  
 * 7:                                   3,7
 * 8:                                   3,8
 * 
 * If this edge has mirrored level, the non-mirrored level will be only populated its edgeInfo
 * Example: Product level has been mirrored
 * edgeInfo: (start, end) 
 * 		Country		    City 		    Product
 * ============================================ 
 * 0: 	-1,0 			0,0 	       [P1,P2,P3]
 * 1: 	-1,2 			0,3 			
 * 2:	 				1,6 			
 * 3: 					1,7 	
 * The product level's value will be sorted according to its basic sort definition. But in case 
 * of aggregation sort, we should try to keep its original sort result.
 * 
 */

public class EdgeInfoGenerator
{
	
	// edge info vector array, to keep all levels edge information
	private Vector[] edgeInfoArray;
	// result set for this edge
	private IAggregationResultSet rs;
	// the dimension axis on this edge
	private DimensionAxis[] dimAxis;
	// the dimension cursor position for each dimension axis
	private int[] dimensionCursorPosition;
	// certain length for mirrored level, 0 length for non-mirrored level
	private int[] length;
	private int edgeCursor, edgeLength, mirrorStartPosition;
	// result set fetcher util
	private ResultSetFetcher fetcher;

	/**
	 * 
	 * @param resultSet
	 * @param axis
	 */
	public EdgeInfoGenerator( IAggregationResultSet resultSet,
			DimensionAxis[] axis, int mirrorStartPosition )
	{
		assert resultSet != null;
		if ( axis.length == 0 )
			return;
		
		this.rs = resultSet;
		this.dimAxis = axis;
		this.dimensionCursorPosition = new int[axis.length];
		this.mirrorStartPosition = mirrorStartPosition;
		fetcher = new ResultSetFetcher( resultSet );
		initDimensionPosition( );
		edgeCursor = -1;
	}

	/**
	 * Populate edgeInfo, EdgeInfo represents the startPosition and endPosition
	 * to its children. Here, it will distinguish the non-mirrored level and
	 * mirrored level. Only non-mirrored level will generate its EdgeInfo
	 * 
	 * @param isPage
	 * @throws IOException
	 */
	public void populateEdgeInfo( boolean isPage ) throws IOException
	{
		int customDimSize = 0;
		if ( mirrorStartPosition > 0 )
		{
			customDimSize = mirrorStartPosition;
		}
		else
		{
			customDimSize = this.dimAxis.length;
		}

		edgeInfoArray = new Vector[customDimSize];
		length = new int[this.dimAxis.length];

		for ( int i = 0; i < customDimSize; i++ )
		{
			edgeInfoArray[i] = new Vector( );
		}

		Object[] preValue = new Object[customDimSize];
		Object[] currValue = new Object[customDimSize];
		for ( int rowId = 0; rowId < this.rs.length( ); rowId++ )
		{
			rs.seek( rowId );
			if ( !isPage )
				for ( int i = 0; i < customDimSize; i++ )
				{
					// TODO Default use 0 index as level key
					currValue[i] = fetcher.getLevelKeyValue( dimAxis[i].getLevelIndex( ) )[fetcher.getAggrResultSet( )
							.getLevelKeyColCount( dimAxis[i].getLevelIndex( ) ) - 1];
				}
			int breakLevel;
			if ( rowId == 0 || isPage )
				breakLevel = 0;
			else
				breakLevel = getBreakLevel( currValue, preValue, rowId );

			for ( int level = breakLevel; level < customDimSize; level++ )
			{
				EdgeInfo edge = new EdgeInfo( );

				if ( level != 0 )
					edge.startPostion = edgeInfoArray[level - 1].size( ) - 1;
				if ( level == edgeInfoArray.length - 1 )
				{
					edge.endPosition = rowId;
				}
				else
				{
					edge.endPosition = edgeInfoArray[level + 1].size( );
				}
				edgeInfoArray[level].add( edge );
			}
			for ( int i = 0; i < customDimSize; i++ )
			{
				preValue[i] = currValue[i];
			}
		}

		// calculate the edgeCursor length
		if ( mirrorStartPosition > 0 )
		{
			this.edgeLength = this.edgeInfoArray[mirrorStartPosition - 1].size( );
			for ( int i = mirrorStartPosition; i < this.dimAxis.length; i++ )
			{
				length[i] = this.dimAxis[i].getDisctinctValue( ).size( );
				this.edgeLength = this.edgeLength * length[i];
			}
		}
		else
		{
			this.edgeLength = this.rs.length( );
		}
	}

	/**
	 * Move certain dimension cursor to the next row.Return false if the next
	 * row does not exist.
	 * 
	 * @param dimAxisIndex certain dimension cursor
	 * @return
	 * @throws OLAPException
	 */
	public boolean dim_next( int dimAxisIndex ) throws OLAPException
	{
		if ( hasNext( dimAxisIndex ) )
		{
			int position = this.dimensionCursorPosition[dimAxisIndex];
			position++;
			this.dimensionCursorPosition[dimAxisIndex] = position;
			return true;
		}
		else
		{
			return false;
		}
	}
	
	/**
	 * Check whether the cursor on dimAxisIndex has next row.
	 * 
	 * @param dimAxisIndex
	 * @return
	 */
	private boolean hasNext( int dimAxisIndex )
	{
		if ( !this.dimAxis[dimAxisIndex].isMirrored( ) )
		{
			EdgeInfo currentEdgeInfo = findEdgeInfo( dimAxisIndex );
			if ( currentEdgeInfo == null && rs.length( ) > 0 )
				return true;
			if ( currentEdgeInfo == null && rs.length( ) == 0 )
				return false;

			int index = this.edgeInfoArray[dimAxisIndex].indexOf( currentEdgeInfo );
			EdgeInfo nextEdgeInfo = null;
			if ( this.edgeInfoArray[dimAxisIndex].size( ) > index + 1 )
				nextEdgeInfo = (EdgeInfo) this.edgeInfoArray[dimAxisIndex].get( index + 1 );

			if ( nextEdgeInfo == null )
				return false;

			if ( currentEdgeInfo.startPostion == nextEdgeInfo.startPostion )
				return true;
			return false;
		}
		else
		{
			if ( this.dimensionCursorPosition[dimAxisIndex] < length[dimAxisIndex] - 1 )
				return true;
			else
				return false;
		}
	}

	/**
	 * 
	 * @param dimAxisIndex
	 * @return
	 * @throws OLAPException
	 */
	public boolean dim_previous( int dimAxisIndex ) throws OLAPException
	{
		int position = this.dimensionCursorPosition[dimAxisIndex];
		position--;

		if ( position >= 0 )
		{
			this.dimensionCursorPosition[dimAxisIndex] = position;
			return true;
		}
		else
		{
			this.dimensionCursorPosition[dimAxisIndex] = -1;
			return false;
		}
	}
	
	/**
	 * 
	 * @param offset
	 * @param dimAxisIndex
	 * @return
	 * @throws OLAPException
	 */
	public boolean dim_relative( int offset, int dimAxisIndex )
			throws OLAPException
	{
		int index = offset;
		
		if ( offset > 0 )
			for ( ; index != 0; index-- )
			{
				if ( this.dim_next( dimAxisIndex ) )
					continue;
				else
					break;
			}
		else
			for ( ; index != 0; index++ )
			{
				if ( this.dim_previous( dimAxisIndex ) )
					continue;
				else
					break;
			}

		if ( index == 0 )
			return true;
		else
			return false;
	}
	
	/**
	 * 
	 * @param dimAxisIndex
	 * @return
	 */
	public boolean dim_first( int dimAxisIndex )
	{
		if ( this.edgeLength > 0 )
		{
			this.dimensionCursorPosition[dimAxisIndex] = 0;
			return true;
		}
		else
		{
			return false;
		}
	}
	
	/**
	 * 
	 * @param dimAxisIndex
	 * @return
	 */
	public boolean dim_last( int dimAxisIndex )
	{
		if ( !this.dimAxis[dimAxisIndex].isMirrored( ) )
		{
			int range = findCurentDimensionRange( dimAxisIndex );

			if ( this.edgeLength > 0 )
			{
				this.dimensionCursorPosition[dimAxisIndex] = range;
				return true;
			}
			else
			{
				return false;
			}
		}
		else
		{
			if ( this.edgeLength > 0 )
			{
				this.dimensionCursorPosition[dimAxisIndex] = length[dimAxisIndex] - 1;
				return true;
			}
			else
			{
				return false;
			}
		}
	}
	
	/**
	 * 
	 * @param dimAxisIndex
	 * @return
	 */
	public boolean dim_isBeforeFirst( int dimAxisIndex )
	{
		return this.dimensionCursorPosition[dimAxisIndex] < 0;
	}

	/**
	 * 
	 * @param dimAxisIndex
	 * @return
	 */
	public boolean dim_isAfterLast( int dimAxisIndex )
	{
		return this.dimensionCursorPosition[dimAxisIndex] == this.getLastDiemsionLength( )[dimAxisIndex];
	}

	/**
	 * 
	 * @param dimAxisIndex
	 * @return
	 */
	public boolean dim_isFirst( int dimAxisIndex )
	{
		return this.dimensionCursorPosition[dimAxisIndex] == 0;
	}

	/**
	 * 
	 * @param dimAxisIndex
	 * @return
	 */
	public boolean dim_isLast( int dimAxisIndex )
	{
		return this.dimensionCursorPosition[dimAxisIndex] == this.getLastDiemsionLength( )[dimAxisIndex] - 1;
	}

	/**
	 * 
	 * @param dimAxisIndex
	 */
	public void dim_afterLast( int dimAxisIndex )
	{
		this.dimensionCursorPosition[dimAxisIndex] = this.getLastDiemsionLength( )[dimAxisIndex];
	}

	/**
	 * 
	 * @param dimAxisIndex
	 */
	public void dim_beforeFirst( int dimAxisIndex )
	{
		this.dimensionCursorPosition[dimAxisIndex] = -1;
	}

	/**
	 * 
	 * @param dimAxisIndex
	 * @param position
	 */
	public void dim_setPosition( int dimAxisIndex, long position )
	{
		this.dimensionCursorPosition[dimAxisIndex] = (int) position;
	}
	
	/**
	 * 
	 * @param dimAxisIndex
	 * @return
	 */
	public long dim_getPosition( int dimAxisIndex )
	{
		return this.dimensionCursorPosition[dimAxisIndex];
	}
	
	/**
	 * 
	 * @param dimAxisIndex
	 * @param attr
	 * @return
	 * @throws OLAPException 
	 */
	public Object dim_getCurrentMember( int dimAxisIndex, int attr, int sortType )
			throws OLAPException
	{
		if ( !this.dimAxis[dimAxisIndex].isMirrored( ) )
		{
			try
			{
				int position = getCurrentRowPosition( dimAxisIndex );
				if ( position == -1 )
				{
					throw new OLAPException( ResourceConstants.RD_EXPR_RESULT_SET_NOT_START );
				}
				rs.seek( position );
			}
			catch ( IOException e )
			{
				throw new OLAPException( e.getLocalizedMessage( ) );
			}
			return this.fetcher.getValue( dimAxis[dimAxisIndex].getLevelIndex( ),
					attr );
		}
		else
		{
			return fetchValueFromMirror( dimAxisIndex, sortType );
		}
	}

	/**
	 * 
	 * @param dimAxisIndex
	 * @param attrName
	 * @return
	 * @throws OLAPException
	 */
	public Object dim_getCurrentMember( int dimAxisIndex, String attrName,
			int sortType ) throws OLAPException
	{

		if ( !this.dimAxis[dimAxisIndex].isMirrored( ) )
		{
			try
			{
				int position = getCurrentRowPosition( dimAxisIndex );
				if ( position == -1 )
				{
					throw new OLAPException( ResourceConstants.RD_EXPR_RESULT_SET_NOT_START );
				}
				rs.seek( position );
			}
			catch ( IOException e )
			{
				throw new OLAPException( e.getLocalizedMessage( ) );
			}
			int attrIndex = fetcher.getAttributeIndex( dimAxis[dimAxisIndex].getLevelIndex( ),
					attrName );
			if ( attrIndex == -1 )
				throw new OLAPException( ResourceConstants.INVALID_LEVEL_ATTRIBUTE + attrName );
			return fetcher.getValue( dimAxis[dimAxisIndex].getLevelIndex( ),
					attrIndex );
		}
		else
		{
			return fetchValueFromMirror( dimAxisIndex, sortType );
		}
	}

	/**
	 * 
	 * @param dimAxisIndex
	 * @param sortType
	 * @return
	 * @throws OLAPException
	 */
	private Object fetchValueFromMirror( int dimAxisIndex, int sortType )
			throws OLAPException
	{
		if ( this.dimensionCursorPosition[dimAxisIndex] < 0
				|| this.dimensionCursorPosition[dimAxisIndex] >= length[dimAxisIndex] )
			throw new OLAPException( ResourceConstants.RD_EXPR_RESULT_SET_NOT_START );

		if ( sortType == IDimensionSortDefn.SORT_UNDEFINED )
		{
			return this.dimAxis[dimAxisIndex].getDisctinctValue( )
					.get( this.dimensionCursorPosition[dimAxisIndex] );
		}
		
		//find the result if there is aggregation sort definition
		Collection collection = null;
		try
		{
			collection = fetchValueCollectionInEdgeInfo( dimAxisIndex );
		}
		catch ( IOException e )
		{
		}

		Vector v = this.dimAxis[dimAxisIndex].getDisctinctValue( );
		v.removeAll( collection );
		if ( sortType == IDimensionSortDefn.SORT_ASC )
		{
			for ( int i = 0, startSize = v.size( ); i < collection.size( ); i++ )
			{
				v.insertElementAt( collection.iterator( ).next( ), startSize );
				startSize++;
			}
		}
		else
		{
			Iterator iter = collection.iterator( );
			int index = 0;
			while ( iter.hasNext( ) )
			{
				v.insertElementAt( iter.next( ), index );
				index++;
			}
		}
		return v.get( this.dimensionCursorPosition[dimAxisIndex] );
	}

	/**
	 * 
	 * @param dimAxisIndex
	 * @return
	 * @throws IOException
	 */
	private Collection fetchValueCollectionInEdgeInfo( int dimAxisIndex )
			throws IOException
	{
		Set value = new HashSet( );
		int edgeStart = this.getEdgeStart( this.mirrorStartPosition - 1 );
		int edgeEnd = this.getEdgeEnd( this.mirrorStartPosition - 1 );
		while ( edgeStart <= edgeEnd )
		{
			this.rs.seek( edgeStart );
			value.add( this.rs.getLevelKeyValue( dimAxisIndex )[this.rs.getLevelKeyColCount( dimAxisIndex ) - 1] );
			edgeStart++;
		}
		return value;
	}

	/**
	 * 
	 * @param dimAxisIndex
	 * @return
	 */
	private int getCurrentRowPosition( int dimAxisIndex )
	{
		EdgeInfo currentEdgeInfo = this.findEdgeInfo( dimAxisIndex );
		if ( currentEdgeInfo == null )
			return -1;
		EdgeInfo outerMost = findOuterMostChildEdgeInfo( dimAxisIndex,
				currentEdgeInfo );

		return outerMost.endPosition;
	}
	
	/**
	 * Based on current dimensionCursor postion, get current edgeinfo for
	 * certain dimensionAxis.
	 * 
	 * @param dimensionAxis
	 * @return
	 */
	private EdgeInfo findEdgeInfo( int dimensionAxis )
	{
		if ( dimensionAxis < 0 || dimensionAxis > this.dimAxis.length )
		{
			return null;
		}

		EdgeInfo tempEdgeInfo = null, edgeInfo = null;
		int endPostion = 0;
		for ( int index = 0; index <= dimensionAxis; index++ )
		{
			if ( dimensionAxis == index )
			{
				if ( this.edgeInfoArray[index].size( ) > endPostion
						&& this.dimensionCursorPosition[index] > -1
						&& this.dimensionCursorPosition[index] + endPostion < this.edgeInfoArray[index].size( ) )
					edgeInfo = (EdgeInfo) this.edgeInfoArray[index].get( this.dimensionCursorPosition[index]
							+ endPostion );
				else if ( this.dimensionCursorPosition[index] == -1 )
					return null;
				break;
			}
			else
			{
				if ( this.dimensionCursorPosition[index] + endPostion < this.edgeInfoArray[index].size( )
						&& this.dimensionCursorPosition[index] > -1
						&& this.edgeInfoArray[index].size( ) > endPostion )
					tempEdgeInfo = (EdgeInfo) this.edgeInfoArray[index].get( this.dimensionCursorPosition[index]
							+ endPostion );
				else
					return null;
				if ( tempEdgeInfo != null )
				{
					endPostion = tempEdgeInfo.endPosition;;
				}
			}
		}
		return edgeInfo;
	}

	/**
	 * 
	 * @param dimensionAxis
	 * @return
	 */
	private int findCurentDimensionRange( int dimensionAxis )
	{
		if ( dimensionAxis < 0 || dimensionAxis > this.dimAxis.length )
		{
			return 0;
		}

		if ( dimensionAxis == 0 )
		{
			if ( this.dimensionCursorPosition[0] < this.edgeInfoArray[0].size( ) )
				return this.edgeInfoArray[0].size( ) - 1;
			else
				return 0;
		}
		int range = 0;
		EdgeInfo info = this.findEdgeInfo( dimensionAxis );
		if ( info == null )
			return range;
		else
		{
			range = 0;
			EdgeInfo currentInfo = info;
			while ( true )
			{
				int index = this.edgeInfoArray[dimensionAxis].indexOf( currentInfo );
				EdgeInfo nextEdgeInfo = null;
				if ( this.edgeInfoArray[dimensionAxis].size( ) > index + 1 )
					nextEdgeInfo = (EdgeInfo) this.edgeInfoArray[dimensionAxis].get( index + 1 );
				else
					break;
				if ( nextEdgeInfo != null
						&& currentInfo.startPostion == nextEdgeInfo.startPostion )
				{
					range++;
					currentInfo = nextEdgeInfo;
				}
				else
				{
					break;
				}
			}

			currentInfo = info;
			while ( true )
			{
				int index = this.edgeInfoArray[dimensionAxis].indexOf( currentInfo );
				EdgeInfo preEdgeInfo = null;
				if ( index > 0 )
					preEdgeInfo = (EdgeInfo) this.edgeInfoArray[dimensionAxis].get( index - 1 );
				else
					break;
				if ( preEdgeInfo != null
						&& currentInfo.startPostion == preEdgeInfo.startPostion )
				{
					range++;
					currentInfo = preEdgeInfo;
				}
				else
				{
					break;
				}
			}

		}

		return range;
	}

	/**
	 * 
	 */
	public void edge_afterLast( )
	{
		this.edgeCursor = edgeLength;
		int[] lastDimLength = this.getLastDiemsionLength( );
		for ( int i = 0; i < this.dimAxis.length; i++ )
		{
			this.dim_setPosition( i, lastDimLength[i] );
		}
	}

	/**
	 * 
	 */
	public void edge_beforeFirst( )
	{
		this.edgeCursor = -1;
		this.initDimensionPosition( );
	}

	/**
	 * 
	 * @return
	 */
	public boolean edge_first( )
	{
		if ( edgeLength > 0 )
		{
			this.edgeCursor = 0;
			for ( int i = 0; i < this.dimAxis.length; i++ )
			{
				this.dimensionCursorPosition[i] = 0;
			}
			return true;
		}
		else
		{
			return false;
		}
	}

	/**
	 * 
	 * @return
	 */
	public long getEdgePostion( )
	{
		return this.edgeCursor;
	}

	/**
	 * 
	 * @return
	 */
	public boolean edge_isAfterLast( )
	{
		return edgeCursor >= edgeLength;
	}

	/**
	 * 
	 * @return
	 */
	public boolean edge_isBeforeFirst( )
	{
		return edgeCursor < 0;
	}

	/**
	 * 
	 * @return
	 */
	public boolean edge_isFirst( )
	{
		return edgeCursor == 0;
	}

	/**
	 * 
	 * @return
	 */
	public boolean edge_isLast( )
	{
		return edgeCursor == edgeLength - 1;
	}

	/**
	 * 
	 * @return
	 */
	public boolean edge_last( )
	{
		if ( edgeLength > 0 )
		{
			edgeCursor = edgeLength - 1;
			int[] lastDimLength = getLastDiemsionLength( );
			for ( int i = 0; i < this.dimAxis.length; i++ )
			{
				this.dim_setPosition( i, lastDimLength[i] - 1 );
			}
			return true;
		}
		else
		{
			return false;
		}
	}

	/**
	 * 
	 * @return
	 * @throws OLAPException
	 */
	public boolean edge_next( ) throws OLAPException
	{
		edgeCursor++;
		if ( edgeCursor < edgeLength )
		{
			if ( this.isInitialStatus( ) )
			{
				for ( int i = this.dimAxis.length - 1; i >= 0; i-- )
				{
					this.dimensionCursorPosition[i] = 0;
				}
			}
			else
				for ( int i = this.dimAxis.length - 1; i >= 0; i-- )
				{
					if ( this.dim_next( i ) )
					{
						break;
					}
					else
					{
						this.dim_first( i );
					}
				}
			return true;
		}
		else
		{
			this.edge_afterLast( );
			return false;
		}
	}

	private boolean isInitialStatus( )
	{
		for ( int i = 0; i < this.dimensionCursorPosition.length; i++ )
		{
			if ( this.dimensionCursorPosition[i] >= 0 )
			{
				return false;
			}
		}
		return true;
	}

	/**
	 * 
	 * @return
	 * @throws OLAPException
	 */
	public boolean edge_previous( ) throws OLAPException
	{
		edgeCursor--;
		if ( edgeCursor >= this.edgeLength - 1 )
		{
			for ( int i = 0; i < this.dimAxis.length; i++ )
			{
				this.dim_previous( i );
			}
			return true;
		}
		else if ( edgeCursor >= 0 )
		{
			for ( int i = this.dimAxis.length - 1; i >= 0; i-- )
			{
				if ( this.dim_previous( i ) )
				{
					for ( int k = i + 1; k < this.dimAxis.length; k++ )
					{
						this.dim_last( k );
					}
					break;
				}
				else
				{
					this.dim_first( i );
				}
			}
			return true;
		}
		else
		{
			edgeCursor = -1;
			this.initDimensionPosition( );
			return false;
		}
	}

	/**
	 * 
	 * @param arg0
	 * @return
	 * @throws OLAPException 
	 */
	public boolean edge_relative( int arg0 ) throws OLAPException
	{
		if ( arg0 == 0 )
			return true;
		int position = edgeCursor + arg0;
		if ( position >= edgeLength )
		{
			this.edge_afterLast( );
			return false;
		}
		else if ( position < 0 )
		{
			this.initDimensionPosition( );
			edgeCursor = -1;
			return false;
		}
		else
		{
			for ( int i = 0; i < Math.abs( arg0 ); i++ )
			{
				if ( arg0 > 0 )
					this.edge_next( );
				else
					this.edge_previous( );
			}
			return true;
		}
	}

	/**
	 * 
	 * @param position
	 * @throws OLAPException 
	 * @throws OLAPException
	 */
	public void edge_setPostion( long position ) throws OLAPException
	{
		if ( position < 0 )
		{
			this.initDimensionPosition( );
			this.edgeCursor = -1;
			return;
		}
		int offSet = (int) position - edgeCursor;
		this.edge_relative( offSet );
	}

	/**
	 * 
	 * @param dimAxisIndex
	 * @return
	 */
	public int getEdgeStart( int dimAxisIndex )
	{

		if ( this.mirrorStartPosition == 0 )
		{
			int segment = this.edgeInfoArray[this.dimAxis.length - 1].size( );
			if ( segment == 1 )
			{
				return this.rs.length( ) - 1;
			}
			EdgeInfo edgeInfo = findEdgeInfo( dimAxisIndex );
			if ( edgeInfo == null )
				return -1;
			int endPosition = edgeInfo.endPosition, position;
			EdgeInfo info = edgeInfo;
			for ( position = dimAxisIndex + 1; position < this.dimAxis.length; position++ )
			{
				if ( this.dimAxis[position].isMirrored( ) )
					break;
				info = (EdgeInfo) ( (List) this.edgeInfoArray[position] ).get( endPosition );
				endPosition = info.endPosition;
			}
			if ( position == this.dimAxis.length )
			{
				return info.endPosition;
			}
			else
			{
				return this.edgeInfoArray[position - 1].indexOf( info )
						* length[position]
						+ this.dimensionCursorPosition[position];
			}
		}
		else
		{
			if ( !this.dimAxis[dimAxisIndex].isMirrored( ) )
			{
				EdgeInfo edgeInfo = findEdgeInfo( dimAxisIndex );
				int start = findOuterMostChildEdgeInfoIndex( dimAxisIndex,
						edgeInfo );
				if ( mirrorStartPosition > 0 && start >= 0 )
					for ( int i = mirrorStartPosition; i < this.dimAxis.length; i++ )
					{
						start = start * length[i];
					}
				return start;
			}
			else
			{
				int start = caculateOffset( dimAxisIndex , true );
				if( start <0 )
					return start;
				if ( dimAxisIndex == this.dimAxis.length - 1 )
				{
					start += this.dimensionCursorPosition[dimAxisIndex];
				}
				return start;
			}
		}
	}

	/**
	 * 
	 * @param dimAxisIndex
	 * @return
	 */
	public int getEdgeEnd( int dimAxisIndex )
	{
		if ( this.mirrorStartPosition == 0 )
		{
			int segment = this.edgeInfoArray[this.dimAxis.length - 1].size( );
			if ( segment == 1 )
			{
				return this.rs.length( ) - 1;
			}
			EdgeInfo edgeInfo = findEdgeInfo( dimAxisIndex );
			if ( edgeInfo == null )
				return -1;
			int endPosition = -1;

			int index = this.edgeInfoArray[dimAxisIndex].indexOf( edgeInfo );
			if ( index < this.edgeInfoArray[dimAxisIndex].size( ) - 1 )
			{
				EdgeInfo nextEdgeInfo = (EdgeInfo) this.edgeInfoArray[dimAxisIndex].get( index + 1 );
				EdgeInfo nextOuterEdgeInfo = this.findOuterMostChildEdgeInfo( dimAxisIndex,
						nextEdgeInfo );
				endPosition = nextOuterEdgeInfo.endPosition - 1;
			}
			else
			{
				endPosition = this.rs.length( ) - 1;
			}
			return endPosition;
		}
		else
		{
			if ( !this.dimAxis[dimAxisIndex].isMirrored( ) )
			{
				EdgeInfo edgeInfo = findEdgeInfo( dimAxisIndex );

				int index = this.edgeInfoArray[dimAxisIndex].indexOf( edgeInfo );
				if ( index < this.edgeInfoArray[dimAxisIndex].size( ) - 1 )
				{
					EdgeInfo nextEdgeInfo = (EdgeInfo) this.edgeInfoArray[dimAxisIndex].get( index + 1 );
					int start = findOuterMostChildEdgeInfoIndex( dimAxisIndex,
							nextEdgeInfo );
					if ( mirrorStartPosition > 0 )
						for ( int i = mirrorStartPosition; i < this.dimAxis.length; i++ )
						{
							start = start * length[i];
						}
					return start - 1;
				}
				else
				{
					return this.edgeLength - 1;
				}
			}
			else
			{
				int start = caculateOffset( dimAxisIndex, false );
				if( start <0 )
					return start;
				if ( dimAxisIndex == this.dimAxis.length - 1 )
				{
					start += this.dimensionCursorPosition[dimAxisIndex] + 1;
				}
				return start - 1;
			}
		}
	}

	/**
	 * 
	 * @param dimAxisIndex
	 * @return
	 */
	private int caculateOffset( int dimAxisIndex , boolean isStart )
	{
		EdgeInfo edgeInfo = findEdgeInfo( mirrorStartPosition - 1 );
		int start = findOuterMostChildEdgeInfoIndex( mirrorStartPosition - 1,
				edgeInfo );
		if( start<0 )
			return start;
		for ( int i = mirrorStartPosition; i < this.dimAxis.length; i++ )
		{
			start = start * length[i];
		}
		for ( int k = mirrorStartPosition; k <= dimAxisIndex; k++ )
		{
			int offset = 1;
			for ( int i = k + 1; i < this.dimAxis.length; i++ )
			{
				if ( k == dimAxisIndex && !isStart )
					offset = offset
							* ( this.dimensionCursorPosition[k] + 1 )
							* length[i];
				else
					offset = offset
							* this.dimensionCursorPosition[k]
							* length[i];
			}
			if ( k + 1 < this.dimAxis.length )
				start += offset;
		}
		return start;
	}

	/**
	 * 
	 */
	private void initDimensionPosition( )
	{
		for ( int i = 0; i < dimAxis.length; i++ )
		{
			dimensionCursorPosition[i] = -1;
		}
	}

	/**
	 * 
	 * @param dimAxisIndex
	 * @param edgeInfo
	 * @return
	 */
	private EdgeInfo findOuterMostChildEdgeInfo( int dimAxisIndex,
			EdgeInfo edgeInfo )
	{
		if ( dimAxisIndex < 0
				|| dimAxisIndex >= this.dimAxis.length || edgeInfo == null )
			return null;
		int endPosition = edgeInfo.endPosition;
		EdgeInfo info = edgeInfo;
		for ( int i = dimAxisIndex + 1; i < this.dimAxis.length; i++ )
		{
			if ( this.dimAxis[i].isMirrored( ) )
				break;
			info = (EdgeInfo) ( (List) this.edgeInfoArray[i] ).get( endPosition );
			endPosition = info.endPosition;
		}
		return info;
	}

	/**
	 * 
	 * @param dimAxisIndex
	 * @param edgeInfo
	 * @return
	 */
	private int findOuterMostChildEdgeInfoIndex( int dimAxisIndex,
			EdgeInfo edgeInfo )
	{
		if ( dimAxisIndex < 0
				|| dimAxisIndex >= this.dimAxis.length || edgeInfo == null )
			return -1;
		int endPosition = edgeInfo.endPosition;
		EdgeInfo info = edgeInfo;
		int index;
		for ( index = dimAxisIndex + 1; index < this.dimAxis.length; index++ )
		{
			if ( this.dimAxis[index].isMirrored( ) )
				break;
			info = (EdgeInfo) ( (List) this.edgeInfoArray[index] ).get( endPosition );
			endPosition = info.endPosition;
		}
		if ( index == this.mirrorStartPosition )
			return this.edgeInfoArray[index - 1].indexOf( info );
		else
			return this.edgeInfoArray[index].indexOf( info );
	}
	
	/**
	 * 
	 * @return
	 */
	private int[] getLastDiemsionLength( )
	{
		int[] lastDimensionLength = new int[this.dimAxis.length];
		if ( this.mirrorStartPosition == 0 )
		{
			for ( int i = 0; i < this.dimAxis.length; i++ )
			{
				lastDimensionLength[i] = this.getRangeInLastDimension( i );
			}
		}
		else
		{
			for ( int i = 0; i < this.mirrorStartPosition; i++ )
			{
				lastDimensionLength[i] = this.getRangeInLastDimension( i );
			}
			for ( int i = this.mirrorStartPosition; i < this.dimAxis.length; i++ )
			{
				lastDimensionLength[i] = length[i];
			}
		}
		return lastDimensionLength;
	}

	/**
	 * 
	 * @param dimIndex
	 * @return
	 */
	private int getRangeInLastDimension( int dimIndex )
	{
		if ( dimIndex == 0 )
			return this.edgeInfoArray[0].size( );
		int size = this.edgeInfoArray[dimIndex].size( );
		if( size == 0 )
			return -1;
		int count = 1;
		EdgeInfo edgeInfo = (EdgeInfo) this.edgeInfoArray[dimIndex].get( size - 1 );
		EdgeInfo previousInfo;
		for ( int i = size - 2; i >= 0; i-- )
		{
			previousInfo = (EdgeInfo) this.edgeInfoArray[dimIndex].get( i );
			if ( previousInfo.startPostion == edgeInfo.startPostion )
				count++;
		}
		return count;
	}
	
	/**
	 * 
	 * @param aggregationIndex
	 * @return
	 * @throws IOException
	 */
	public Object dim_getCurrentAggregation( int aggregationIndex )
			throws IOException
	{
		return this.rs.getAggregationValue( aggregationIndex );
	}

	/**
	 * 
	 * @param currValue
	 * @param preValue
	 * @param rowId
	 * @return
	 */
	private int getBreakLevel( Object[] currValue, Object[] preValue, int rowId )
	{
		assert preValue != null && currValue != null;
		int breakLevel = 0;
		for ( ; breakLevel < currValue.length; breakLevel++ )
		{
			// get the first child of current group in level of breakLevel
			List list = this.edgeInfoArray[breakLevel];
			EdgeInfo edgeInfo = (EdgeInfo) list.get( list.size( ) - 1 );
			int child = edgeInfo.endPosition;

			Object currObjectValue = currValue[breakLevel];
			Object prevObjectValue = preValue[breakLevel];

			for ( int level = breakLevel + 1; level < this.edgeInfoArray.length; level++ )
			{
				list = this.edgeInfoArray[level];
				edgeInfo = (EdgeInfo) list.get( child );
				child = edgeInfo.endPosition;
			}

			// determins whether next row is in current group
			if ( isEqualObject( currObjectValue, prevObjectValue ) == false )
			{
				break;
			}

		}
		return breakLevel;
	}

	/**
	 * 
	 * @param preValue
	 * @param currentValue
	 * @return
	 */
	private boolean isEqualObject( Object preValue, Object currentValue )
	{
		if ( preValue == currentValue )
		{
			return true;
		}
		if ( preValue == null || currentValue == null )
		{
			return false;
		}
		return preValue.equals( currentValue );
	}
}

class EdgeInfo
{
	int startPostion = -1;
	int endPosition = -1;
}
