/*******************************************************************************
 * Copyright (c) 2006 Inetsoft Technology Corp. All rights reserved. This
 * program and the accompanying materials are made available under the terms of
 * the Eclipse Public License v1.0 which accompanies this distribution, and is
 * available at http://www.eclipse.org/legal/epl-v10.html Contributors: Inetsoft
 * Technology Corp - Implementation
 ******************************************************************************/

package org.eclipse.birt.report.engine.emitter.wpml;

import java.awt.Image;
import java.io.BufferedOutputStream;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URL;
import java.util.HashSet;
import java.util.List;
import java.util.Stack;
import java.util.Vector;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.eclipse.birt.report.engine.api.IHTMLActionHandler;
import org.eclipse.birt.report.engine.api.RenderOptionBase;
import org.eclipse.birt.report.engine.api.impl.Action;
import org.eclipse.birt.report.engine.content.IAutoTextContent;
import org.eclipse.birt.report.engine.content.IBandContent;
import org.eclipse.birt.report.engine.content.ICellContent;
import org.eclipse.birt.report.engine.content.IColumn;
import org.eclipse.birt.report.engine.content.IContainerContent;
import org.eclipse.birt.report.engine.content.IContent;
import org.eclipse.birt.report.engine.content.IDataContent;
import org.eclipse.birt.report.engine.content.IForeignContent;
import org.eclipse.birt.report.engine.content.IGroupContent;
import org.eclipse.birt.report.engine.content.IHyperlinkAction;
import org.eclipse.birt.report.engine.content.IImageContent;
import org.eclipse.birt.report.engine.content.ILabelContent;
import org.eclipse.birt.report.engine.content.IListBandContent;
import org.eclipse.birt.report.engine.content.IListContent;
import org.eclipse.birt.report.engine.content.IListGroupContent;
import org.eclipse.birt.report.engine.content.IPageContent;
import org.eclipse.birt.report.engine.content.IReportContent;
import org.eclipse.birt.report.engine.content.IRowContent;
import org.eclipse.birt.report.engine.content.IStyle;
import org.eclipse.birt.report.engine.content.ITableBandContent;
import org.eclipse.birt.report.engine.content.ITableContent;
import org.eclipse.birt.report.engine.content.ITableGroupContent;
import org.eclipse.birt.report.engine.content.ITextContent;
import org.eclipse.birt.report.engine.css.engine.value.birt.BIRTConstants;
import org.eclipse.birt.report.engine.emitter.ContentEmitterAdapter;
import org.eclipse.birt.report.engine.emitter.IEmitterServices;
import org.eclipse.birt.report.engine.ir.EngineIRConstants;
import org.eclipse.birt.report.engine.ir.SimpleMasterPageDesign;
import org.eclipse.birt.report.engine.layout.pdf.util.HTML2Content;
import org.eclipse.birt.report.engine.presentation.ContentEmitterVisitor;
import org.w3c.dom.css.CSSValue;

public class WordEmitter extends ContentEmitterAdapter
{

	protected static Logger logger = Logger.getLogger( WordEmitter.class
			.getName( ) );

	protected OutputStream out = null;

	IEmitterServices eservice = null;

	protected ContentEmitterVisitor contentVisitor;

	WpmlWriter writer = null;

	public static final String OUTPUT_FORMAT = "doc";

	private int pageWidth = 0;

	private int pageHeight = 0;

	private int contentHeight = 0;

	private int contentWidth = 0;

	private int headerHeight = 0;

	private int footerHeight = 0;

	private int topMargin = 0;

	private int bottomMargin = 0;

	private int leftMargin = 0;

	private int rightMargin = 0;
	
	private boolean firstBg = true;

	private String orientation = "portrait";

	private EmitterContext context = null;

	private IPageContent page = null;

	private Stack styles = new Stack( );

	private Vector cacheCells = new Vector( );

	final static int NORMAL = -1;

	private static int firstPage = 0;
	
	private static int normalPage = 1;
	
	private int pageStatus;
	
	private HashSet marks = new HashSet();

	public WordEmitter( )
	{
		contentVisitor = new ContentEmitterVisitor( this );
	}

	public void initialize( IEmitterServices service )
	{
		eservice = service;

		if ( service != null )
		{
			Object fd = eservice.getOption( RenderOptionBase.OUTPUT_FILE_NAME );
			File file = null;

			if ( fd != null )
			{
				try
				{
					file = new File( fd.toString( ) );

					File parent = file.getParentFile( );

					if ( parent != null && !parent.exists( ) )
					{
						parent.mkdirs( );
					}
					out = new BufferedOutputStream( new FileOutputStream( file ) );
				}
				catch ( FileNotFoundException e )
				{
					logger.log( Level.WARNING, e.getMessage( ), e );
				}
			}
		}

		if ( out == null )
		{
			Object val = eservice.getOption( RenderOptionBase.OUTPUT_STREAM );

			if ( val != null && val instanceof OutputStream )
			{
				out = (OutputStream) val;
			}
		}

		writer = new WpmlWriter( out );
		context = new EmitterContext( );

		// Assume the out is always setup successfully
	}

	public String getOutputFormat( )
	{
		return WordEmitter.OUTPUT_FORMAT;
	}

	public void startPage( IPageContent page )
	{

		if ( this.page != null )
		{
			completePrePage( this.page );
			pageStatus = WordEmitter.normalPage;
			context.resetWidth( );

		}
		else
		{
			pageStatus = WordEmitter.firstPage;
			writer.start( );
		}

		this.page = page;
		accountPageProp( page );
		context.addWidth( contentWidth );

		writer.startPg( );

	}

	public void completePrePage( IPageContent page )
	{
		if ( context.isInline( ) )
		{
			writer.endParagraph( );
		}

		context.endInline( );

		writer.startPageBreak( );

		writeSection( page );

		writer.endPageBreak( );

		writer.endPg( );
		
		handlePageBg();
	}
	
	/**
	 * Transfer background for current page to Doc format.
	 * Now, the exported file will apply the first background 
	 * properties, and followed background will ignore. 
	 * 
	 * In addition, Since the Word only support fill-in background,
	 * the background attach, pos, posX, posY and repeat are not mapped 
	 * to Word easyly. 
	 * At present, ignore those properties.
	 * 
	 * @TODO support background properties. attach, pos, posx, posy and repeat.  
	 */
	
	private void handlePageBg() 
	{
		// Set the first page background which is not null to DOC
		if ( firstBg )
		{
			IStyle style = page.getStyle( );

			String bgcolor = style.getBackgroundColor( );
			String bgimg = style.getBackgroundImage( );
			
			// Image priority is higher than color.
			if(bgimg != null)
			{
				try {
					File file = new File(bgimg);					
					byte[] img = getImage(new FileInputStream(file));
					String pic2Text = new sun.misc.BASE64Encoder( ).encode( img );
					writer.writePageBackground("image", pic2Text);
					firstBg = false;
				}
				catch(Exception e) {
					logger.log( Level.WARNING, "Can not get background image" );
				}
			}
			else if(bgcolor != null && !"transparent".equalsIgnoreCase( bgcolor ))
			{
				writer.writePageBackground("color", bgcolor);
				firstBg = false;
			}	
			// String attach = style.getBackgroundAttachment( );
			// String pos = style.getBackgroundPosition( );
			// String posX = style.getBackgroundPositionX( );
			// String posY = style.getBackgroundPositionY( );
			// String repeat = style.getBackgroundRepeat( );			
		}
	}

	public void end( IReportContent report )
	{
		if ( context.isInline( ) )
		{
			writer.endParagraph( );
		}
		
		context.endInline( );

		writeSection( page );
		// completePrePage(page, true);
		writer.endPg( );
		
		handlePageBg();
		
		writer.end( );

	}

	public void writeSection( IPageContent page )
	{
		accountPageProp( page );

		writer.startSection( );

		if ( pageStatus == WordEmitter.firstPage )
		{
			writeHeaderFooter( );
		}

		writer.writePgProp(
				pageHeight,
				pageWidth,
				headerHeight,
				footerHeight,
				topMargin,
				bottomMargin,
				leftMargin,
				rightMargin,
				this.orientation );

		writer.endSection( );
	}

	public void writeHeaderFooter( )
	{

		SimpleMasterPageDesign master = (SimpleMasterPageDesign) ( page
				.getReportContent( )
				.getDesign( )
				.getPageSetup( ).getMasterPage( 0 ) );
		if ( !master.isShowHeaderOnFirst( ) )
		{
			writer.startHeader( true );
		}
		else
		{
			writer.startHeader( false );
		}

		if ( page.getPageHeader( ) != null )
		{
			contentVisitor.visitChildren( page.getPageHeader( ), null );
		}
		writer.endHeader( );
		writer.startFooter( );

		if ( page.getPageFooter( ) != null )
		{
			contentVisitor.visitChildren( page.getPageFooter( ), null );
		}
		writer.endFooter( );

	}

	public void accountPageProp( IPageContent page )
	{

		pageWidth = WordUtil.convertTo( page.getPageWidth( ), 0 );
		// 11 inch * 1440
		pageHeight = WordUtil.convertTo( page.getPageHeight( ), 0 );

		footerHeight = WordUtil.convertTo( page.getFooterHeight( ), 0 );
		headerHeight = WordUtil.convertTo( page.getHeaderHeight( ), 0 );

		topMargin = WordUtil.convertTo( page.getMarginTop( ), 0 );
		bottomMargin = WordUtil.convertTo( page.getMarginBottom( ), 0 );

		contentHeight = pageHeight - headerHeight
				- footerHeight - topMargin - bottomMargin;

		leftMargin = WordUtil.convertTo( page.getMarginLeft( ), 0 );
		rightMargin = WordUtil.convertTo( page.getMarginRight( ), 0 );

		contentWidth = pageWidth - leftMargin - rightMargin;

		this.orientation = page.getOrientation( );

	}

	public void endCell( ICellContent cell )
	{
		adjustInline( );
		context.removeWidth( );
		writer.endTblCell( context.needEmptyP( ) );
		context.endCell( );
	}

	public void endContainer( IContainerContent container )
	{
		adjustInline( );

		if ( !styles.isEmpty( ) )
		{
			styles.pop( );
		}

		writer.endTblCell( context.needEmptyP( ) );
		context.endCell( );
		context.addContainer( true );
		writer.endTblRow( );
		writer.endTable( );
		context.setLast( true );
	}

	public void endContent( IContent content )
	{
		// TODO Auto-generated method stub

	}

	public void endGroup( IGroupContent group )
	{
		// TODO Auto-generated method stub

	}

	public void endList( IListContent list )
	{
		if ( !styles.isEmpty( ) )
		{
			styles.pop( );
		}

		context.addContainer( true );
		writer.endTable( );
		context.setLast( true );
	}

	public void endListBand( IListBandContent listBand )
	{
		adjustInline( );
		context.endCell( );
		writer.endTblCell( true );
		writer.endTblRow( );
	}

	public void endListGroup( IListGroupContent group )
	{
		// TODO Auto-generated method stub

	}

	public void endPage( IPageContent page )
	{

	}

	public void endRow( IRowContent row )
	{
		if ( !isHidden( row ) )
		{
			styles.pop( );
			int col = context.getCurrentTblCols( ).length - 1;

			List spans = context.getSpans( col );

			if ( spans != null )
			{
				for ( int i = 0; i < spans.size( ); i++ )
				{
					writer.writeSpanCell( (SpanInfo) spans.get( i ) );
				}
			}

			writer.endTblRow( );
		}		
	}

	public void endTable( ITableContent table )
	{
		context.addContainer( true );

		if ( !styles.isEmpty( ) )
		{
			styles.pop( );
		}

		writer.endTable( );
		context.setLast( true );
		context.removeTable( );
	}

	public void endTableBand( ITableBandContent band )
	{
		// TODO Auto-generated method stub

	}

	public void endTableGroup( ITableGroupContent group )
	{
		// TODO Auto-generated method stub
	}

	public void start( IReportContent report )
	{
	}

	public void startAutoText( IAutoTextContent autoText )
	{
		writeContent( autoText.getType( ), null, autoText );
	}

	public void startCell( ICellContent cell )
	{
		context.startCell( );
		int col = cell.getColumn( );
		List spans = context.getSpans( col );

		if ( spans != null )
		{
			for ( int i = 0; i < spans.size( ); i++ )
			{
				writer.writeSpanCell( (SpanInfo) spans.get( i ) );
			}
		}

		// for(int i = 0; i < cacheCells.size( ); i++)
		// {
		// SpanCell scell = (SpanCell)cacheCells.get( i );
		//			
		// if(scell.cell.getRowSpan( ) > 1)
		// {
		// context.addSpan( scell.col, scell.cell.getColSpan( ), scell.cw,
		// scell.rowSpan, scell.style
		// }
		//			
		// }

		int cs = cell.getColSpan( );
		int rs = cell.getRowSpan( );
		int cw = context.getCellWidth( col, cs );

		IStyle style = computeStyle( cell.getComputedStyle( ) );

		if ( rs > 1 )
		{
			context.addSpan( col, cs, cw, rs, style );
			cacheCells.add( new SpanCell( cell, rs, col, cw, style ) );
		}

		SpanInfo info = null;

		if ( ( cs > 1 ) || ( rs > 1 ) )
		{
			info = new SpanInfo( col, cs, cw, true, style );
		}

		writer.startTblCell( cw, style, info );
		context.addWidth( WordUtil.getWidth( cw, style ) );
	}

	private IStyle computeStyle( IStyle style )
	{
		if ( styles.size( ) == 0 )
		{
			return style;
		}

		for ( int i = 0; i < style.NUMBER_OF_STYLE; i++ )
		{
			if ( isHeritProperty( i ) )
			{
				if ( isNullValue( style.getProperty( i ) ) )
				{
					style.setProperty( i, null );

					for ( int p = styles.size( ) - 1; p >= 0; p-- )
					{
						IStyle pstyle = (IStyle) styles.get( p );

						if ( !isNullValue( pstyle.getProperty( i ) ) )
						{
							style.setProperty( i, pstyle.getProperty( i ) );
							break;
						}
					}
				}
			}
		}

		return style;
	}

	private boolean isNullValue( CSSValue value )
	{
		String[] nullvals = new String[]{"none", "transparent"};

		if ( value == null )
		{
			return true;
		}

		for ( int i = 0; i < nullvals.length; i++ )
		{
			if ( nullvals[i].equalsIgnoreCase( value.getCssText( ) ) )
			{
				return true;
			}
		}

		return false;
	}

	private boolean isHeritProperty( int p )
	{
		int[] ignores = new int[]{IStyle.STYLE_BORDER_BOTTOM_COLOR,
				IStyle.STYLE_BORDER_BOTTOM_STYLE,
				IStyle.STYLE_BORDER_BOTTOM_WIDTH,
				IStyle.STYLE_BORDER_TOP_COLOR, IStyle.STYLE_BORDER_TOP_STYLE,
				IStyle.STYLE_BORDER_TOP_WIDTH, IStyle.STYLE_BORDER_LEFT_COLOR,
				IStyle.STYLE_BORDER_LEFT_STYLE, IStyle.STYLE_BORDER_LEFT_WIDTH,
				IStyle.STYLE_BORDER_RIGHT_COLOR,
				IStyle.STYLE_BORDER_RIGHT_STYLE,
				IStyle.STYLE_BORDER_RIGHT_WIDTH};

		for ( int i = 0; i < ignores.length; i++ )
		{
			if ( p == ignores[i] )
			{
				return false;
			}
		}

		return true;
	}

	public void startContainer( IContainerContent container )
	{
		adjustInline( );

		context.startCell( );

		if(context.isLastTable( ))
		{
			writer.insertHiddenP( );
		}
		
		writer.startTable( container.getComputedStyle( ), context
				.getCurrentWidth( ) );
		writer.startTblRow( -1, "normal" );
		styles.push( container.getComputedStyle( ) );
		IStyle style = computeStyle( container.getComputedStyle( ) );
		writer.startTblCell( context.getCurrentWidth( ), style, null );		
	}

	public void startContent( IContent content )
	{
		// TODO Auto-generated method stub
	}

	public void startData( IDataContent data )
	{
		writeContent( WordEmitter.NORMAL, data.getText( ), data );
	}

	public void startForeign( IForeignContent foreign )
	{
		if ( IForeignContent.HTML_TYPE.equalsIgnoreCase( foreign.getRawType( ) ) )
		{
			HTML2Content convert = new HTML2Content( foreign
					.getReportContent( )
					.getDesign( )
					.getReportDesign( ) );
			convert.html2Content( foreign );

			context.startCell( );
			
			if(context.isLastTable( ))
			{
				writer.insertHiddenP( );
			}
			
			writer.startTable( foreign.getComputedStyle( ), context
					.getCurrentWidth( ) );
			writer.startTblRow( -1, "normal" );
			writer.startTblCell( context.getCurrentWidth( ), foreign
					.getComputedStyle( ), null );

			contentVisitor.visitChildren( foreign, null );

			adjustInline( );
			writer.endTblCell( context.needEmptyP( ) );
			context.endCell( );
			writer.endTblRow( );
			writer.endTable( );
			context.setLast( true );
			context.addContainer( true );
		}
		else
		{
			writeContent(
					WordEmitter.NORMAL,
					foreign.getRawValue( ).toString( ),
					foreign );
		}
	}

	public void startGroup( IGroupContent group )
	{
		// TODO Auto-generated method stub

	}

	public void startImage( IImageContent image )
	{
		adjustInline( );

		byte[] data = parseImage( image );

		if ( data == null )
			return;

		int defaultW = 0;
		int defaultH = 0;
		try
		{
			Image i = javax.imageio.ImageIO.read( new ByteArrayInputStream(
					data ) );

			defaultW = i.getWidth( null );// pix

			defaultH = i.getHeight( null );

		}
		catch ( Exception e )
		{
			logger.log( Level.WARNING, e.getMessage( ), e );
		}

		double height = WordUtil
				.convertImageSize( image.getHeight( ), defaultH );

		double width = WordUtil.convertImageSize( image.getWidth( ), defaultW );

		// recalculate the width ,height according to the cellwidth
		double originalRatio = height / width;

		// //1 inch = 25.4mm 72pt = 1 inch
		// pt is absolute measure different from relative measure px

		double cellWidth = WordUtil.twipToPt( context.getCurrentWidth( ) );

		if ( width > cellWidth )
		{
			width = cellWidth;

			height = (int) ( width * originalRatio );

		}

		String pic2Text = new sun.misc.BASE64Encoder( ).encode( data );

		writeBookmark( image );
		writeToc( image );
		HyperlinkInfo hyper = getHyperlink( image );
		
		
		writer.drawImage( pic2Text, height, width, hyper, image.getComputedStyle( ) );
	}

	public void startLabel( ILabelContent label )
	{
		String txt = label.getText( ) == null ? label.getLabelText( ) : label
				.getText( );
		txt = txt == null ? "" : txt;
		writeContent( WordEmitter.NORMAL, txt, label );
	}

	private Object getReportContext( )
	{
		if ( eservice != null )
		{
			return eservice.getReportContext( );
		}
		return null;
	}

	public void startList( IListContent list )
	{
		adjustInline( );

		styles.push( list.getComputedStyle( ) );

		writeBookmark( list );
		writeToc( list );
		
		if(context.isLastTable( ))
		{
			writer.insertHiddenP( );
		}	
		
		writer
				.startTable( list.getComputedStyle( ), context
						.getCurrentWidth( ) );
	}

	public void startListBand( IListBandContent listBand )
	{
		context.startCell( );
		writer.startTblRow( -1, "normal" );

		IStyle style = computeStyle( listBand.getComputedStyle( ) );

		writer.startTblCell( context.getCurrentWidth( ), style, null );

	}

	public void startListGroup( IListGroupContent group )
	{
		// TODO Auto-generated method stub

	}

	public void startRow( IRowContent row )
	{
		if ( !isHidden( row ) )
		{
			String type = "normal";
			styles.push( row.getComputedStyle( ) );
			if ( row.getBand( ) != null
					&& row.getBand( ).getBandType( ) == IBandContent.BAND_HEADER )
			{
				type = "header";
			}

			double height = WordUtil.convertTo( row.getHeight( ) );

			writer.startTblRow( height, type );

			context.newRow( );
		}
	}

	protected boolean isHidden( IContent content )
	{
		if ( content != null )
		{
			IStyle style = content.getStyle( );
			if(!IStyle.NONE_VALUE.equals(style.getProperty( IStyle.STYLE_DISPLAY )))
			{
				return isHiddenByVisibility( content );
			}
			return true;
		}
		return false;
	}

	/**
	 * if the content is hidden
	 * 
	 * @return
	 */
	private boolean isHiddenByVisibility( IContent content )
	{
		assert content != null;
		IStyle style = content.getStyle( );
		String formats = style.getVisibleFormat( );
		if ( contains(formats, OUTPUT_FORMAT))
		{
			return true;
		}
		return false;
	}
	
	private boolean contains(String formats, String format)
	{
		if ( formats != null
				&& ( formats.indexOf( EngineIRConstants.FORMAT_TYPE_VIEWER ) >= 0
						|| formats.indexOf( BIRTConstants.BIRT_ALL_VALUE ) >= 0 || formats
						.indexOf( format ) >= 0 ) )
		{
			return true;
		}
		return false;
	}
	

	public void startTable( ITableContent table )
	{
		adjustInline( );

		styles.push( table.getComputedStyle( ) );

		IStyle style = table.getComputedStyle( );

		writeBookmark( table );
		writeToc( table );
		
		if(context.isLastTable( ))
		{
			writer.insertHiddenP( );
		}	
		
		writer
				.startTable( table.getComputedStyle( ), context
						.getCurrentWidth( ) );

		int tblWid = WordUtil.convertTo( table.getWidth( ), context
				.getCurrentWidth( ) );
		int colCount = table.getColumnCount( );
		int[] cols = new int[colCount];
		int count = 0;
		int total = 0;

		for ( int i = 0; i < colCount; i++ )
		{
			IColumn col = table.getColumn( i );
			if ( col.getWidth( ) == null )
			{
				cols[i] = -1;
				count++;
			}
			else
			{
				cols[i] = WordUtil.convertTo( col.getWidth( ), tblWid );
				total += cols[i];
			}
		}

		for ( int i = 0; i < cols.length; i++ )
		{
			if ( cols[i] == -1 )
			{
				cols[i] = ( tblWid - total ) / count;
			}
		}

		writer.writeCol( cols );

		context.addTable( cols, style );
	}

	public void startTableBand( ITableBandContent band )
	{

	}

	public void startTableGroup( ITableGroupContent group )
	{
		// TODO Auto-generated method stub

	}

	public void startText( ITextContent text )
	{
		writeContent( WordEmitter.NORMAL, text.getText( ), text );
	}

	// bookmark must not contain space,word may not process bookmark with space
	private void writeToc( IContent content )
	{
		Object tocObj = content.getTOC( );

		if ( tocObj == null )
		{
			return;
		}
		// current only toc with String type can be handled
		if ( tocObj instanceof String )
		{
			String toc = (String) tocObj;
			writer.writeBookmark( toc );
		}
	}

	private void writeBookmark( IContent content )
	{
		Object design = content.getGenerateBy( );
		
		if(marks.contains( design ))
		{
			return;
		}
		
		String bookmark = content.getBookmark( );
		// birt use __TOC_X_X as bookmark for toc and thus it is not a really
		// bookmark
		if ( bookmark == null || bookmark.toLowerCase( ).indexOf( "_toc" ) > 0 )
		{	
			return;
		}
		
		bookmark = bookmark.replaceAll( " ", "_" );
		writer.writeBookmark( bookmark );
		marks.add( design );
	}

	private HyperlinkInfo getHyperlink( IContent content )
	{
		IHyperlinkAction linkaction = content.getHyperlinkAction( );

		HyperlinkInfo hyper = null;

		if ( linkaction != null )
		{
			if ( linkaction.getType( ) == IHyperlinkAction.ACTION_BOOKMARK )
			{
				String bookmark = linkaction.getBookmark( ).replaceAll(
						" ",
						"_" );
				hyper = new HyperlinkInfo( HyperlinkInfo.BOOKMARK, bookmark );
			}
			else if ( linkaction.getType( ) == IHyperlinkAction.ACTION_HYPERLINK )
			{
				hyper = new HyperlinkInfo( HyperlinkInfo.HYPERLINK, linkaction
						.getHyperlink( ) );
			}
			else if ( linkaction.getType( ) == IHyperlinkAction.ACTION_DRILLTHROUGH )
			{
				Action act = new Action( linkaction );
				IHTMLActionHandler actionHandler = null;
				Object ac = eservice
						.getOption( RenderOptionBase.ACTION_HANDLER );

				if ( ac != null && ac instanceof IHTMLActionHandler )
				{
					actionHandler = (IHTMLActionHandler) ac;
					String url = actionHandler
							.getURL( act, getReportContext( ) );
					hyper = new HyperlinkInfo( HyperlinkInfo.HYPERLINK, url );
				}
			}
		}
		return hyper;
	}

	private byte[] getImage( InputStream is ) throws Exception
	{
		ByteArrayOutputStream baos = new ByteArrayOutputStream( );

		int cnt;
		byte[] buf = new byte[4096];

		while ( ( cnt = is.read( buf ) ) >= 0 )
		{
			baos.write( buf, 0, cnt );
		}
		is.close( );
		baos.close( );
		return baos.toByteArray( );

	}

	private byte[] parseImage( IImageContent image )
	{
		byte[] data = null;

		if ( image.getData( ) != null )
		{
			data = image.getData( );
		}
		else
		{
			String url = image.getURI( );

			if ( url == null )
			{
				return null;
			}
			try
			{
				URL Ourl = new URL( url );

				return getImage( Ourl.openStream( ) );
			}
			catch ( Exception e )
			{
				logger.log( Level.WARNING, e.getMessage( ), e );
			}

		}
		return data;
	}

	private void writeContent( int type, String txt, IContent content )
	{
		context.addContainer( false );

		boolean inline = false;
		boolean first_inline = false;

		if ( "inline".equalsIgnoreCase( content
				.getComputedStyle( )
				.getDisplay( ) ) )
		{
			if ( !context.isInline( ) )
			{
				context.startInline( );
				first_inline = true;
			}

			inline = true;
		}
		else
		{
			if ( context.isInline( ) )
			{
				context.endInline( );
				writer.endParagraph( );
			}
		}

		writeBookmark( content );
		writeToc( content ); // element with Toc contains bookmark
		HyperlinkInfo hyper = getHyperlink( content );
		writer.writeContent(
				type,
				txt,
				content.getComputedStyle( ),
				hyper,
				inline,
				first_inline );
		
		context.setLast( false );
	}

	private void adjustInline( )
	{
		if ( context.isInline( ) )
		{
			writer.endParagraph( );
		}

		context.endInline( );
	}

	class SpanCell
	{

		SpanCell( ICellContent cell, int rowSpan, int col, int cw, IStyle style )
		{
			this.cell = cell;
			this.rowSpan = rowSpan;
			this.col = col;
			this.cw = cw;
			this.style = style;
		}

		ICellContent cell;

		int rowSpan, col, cw;

		IStyle style;
	}
}
